from __future__ import annotations
import pydantic

from spice_rack._bases._mixins._mixin_base import PydanticMixinBase
from spice_rack import _guid_service

__all__ = ("GuidMixin", )


class GuidMixin(PydanticMixinBase):
    """
    adds a UUID4 field, 'guid' that is an uuid4 object
    """
    guid: _guid_service.GuidT = pydantic.Field(
        description="globally unique id field",
        default_factory=_guid_service.gen_guid
    )



from __future__ import annotations
import typing as t
import pydantic
import datetime as dt

from spice_rack._ts_service import _validators, _timestamp, _tz_key
from spice_rack import _bases


__all__ = (
    "TimestampAutoParseStrT",
    "timestamp_auto_str_parse_ta",
    "UtcMicrosecondTimestamp",
    "utc_microsecond_timestamp_ta"
)


TimestampAutoParseStrT = t.Annotated[
    _timestamp.Timestamp,
    _validators.StrParser()
]
"""
this type annotation when used within the pydantic context will use dateparser to automatically
parse any strings to the Timestamp instance. This is potentially very slow so use it carefully when convenience 
and flexibility trumps everything else.

Example using as a field annotation::

    class Model(pydantic.BaseModel):
        ts: Timestamp
        ts_flexible: TimestampAutoParseStrT
    
    date_str = "Aug 13th 1995"
    Model(
        ts=date_str,  # will fail validation
        annotated_ts=date_str  # will be parsed
    )
"""


timestamp_auto_str_parse_ta: pydantic.TypeAdapter[TimestampAutoParseStrT] = pydantic.TypeAdapter(TimestampAutoParseStrT)
"""
type adapter to get the pydantic parsing behavior 'TimestampAutoParseStrT' outside the pydantic context.

Example using as standalone parser::

    class Model(pydantic.BaseModel):
        ts: Timestamp
        ts_flexible: TimestampAutoParseStrT
    
    date_str = "Aug 13th 1995"
    Timestamp.model_validate(date_str)  # will fail
    timestamp_auto_str_parse_ta.validate_python(date_str) # will parse to a Timestamp instance
"""


def _utc_us_before_val_func(__obj: t.Any) -> int:
    if isinstance(__obj, (_timestamp.Timestamp, dt.datetime, dt.date)):
        return _timestamp.Timestamp.model_validate(__obj).microseconds
    else:
        return __obj


def _utc_us_after_val_func(__obj: int) -> int:
    # todo: validate it is a valid microseconds from epoch value
    return __obj


_UsIntTypeAnn = t.Annotated[
    int,
    pydantic.BeforeValidator(_utc_us_before_val_func),
    pydantic.AfterValidator(_utc_us_after_val_func)
]


class UtcMicrosecondTimestamp(_bases.RootModel[_UsIntTypeAnn]):
    """
    wraps an integer that passes through our Timestamp class during validation to ensure it is a valid microsecond
    timestamp. This is useful when you want to represent the serialize the timestamp as an int, like in a database
    column or something similar where you don't want the complex object, and you're ok with throwing away
    the timezone information.
    """
    def to_timestamp_inst(self) -> _timestamp.Timestamp:
        return _timestamp.Timestamp(
            microseconds=self.root,
            tz=_tz_key.TimeZoneKey("UTC")
        )


utc_microsecond_timestamp_ta: pydantic.TypeAdapter[int] = pydantic.TypeAdapter(_UsIntTypeAnn)
"""
wraps an integer that passes through our Timestamp class during validation to ensure it is a valid microsecond
timestamp. This is useful when you want to represent the serialize the timestamp as an int, like in a database
column or something similar where you don't want the complex object, and you're ok with throwing away 
the timezone information.
"""



from __future__ import annotations
import typing as t
from typing_extensions import TypeAlias
from io import TextIOWrapper
from fsspec.spec import AbstractBufferedFile


__all__ = (
    "SupportedOpenModesT",
    "OpenFileT"
)


SupportedOpenModesT: TypeAlias = t.Literal["ab", "wb", "rb"]
OpenFileT: TypeAlias = t.Union[AbstractBufferedFile, TextIOWrapper]



from __future__ import annotations
import typing as t
import polars as pl

if t.TYPE_CHECKING:
    from spice_rack._polars_service import _types

__all__ = (
    "HowJoinT",
    "join_dfs",
    "stack_dfs",

)

HowJoinT = t.Literal["inner", "outer", "left"]


def join_dfs(
        dfs: t.Iterable[_types.PolarsMaybeLazyDfT],
        join_on: str,
        how: HowJoinT = "inner"
) -> _types.PolarsLazyDfT:
    """
    join an iterable of polars DataFrame or LazyFrame objects into a single LazyFrame

    Args:
        dfs: iterable of polars objects
        join_on: the column we are joining on, all dataframes must have this column
        how: the type of join, 'inner', 'outer', 'left',
            if 'left' the first dataframe in the iterable is the 'left' one

    Returns:
        pl.LazyFrame: the joined polars LazyFrame
    """
    merged_df: t.Optional[pl.DataFrame] = None
    for df in dfs:

        if isinstance(df, pl.DataFrame):
            df = df.lazy()
        if isinstance(df, pl.LazyFrame):
            df = df.collect()

        if merged_df is None:
            merged_df = df
        else:
            merged_df = merged_df.join(other=df, on=join_on, how=how)

    return merged_df.lazy()


def stack_dfs(
        dfs: t.List[_types.PolarsMaybeLazyDfT],
) -> _types.PolarsLazyDfT:
    """
    vertically stack an iterable of polars DataFrame or LazyFrame objects into a single LazyFrame.
    This is a vertical stack.

    Args:
        dfs: iterable of polars objects

    Returns:
        pl.LazyFrame: the stacked polars LazyFrame
    """

    if len(dfs) == 0:
        raise ValueError("empty list passed in")

    col_sets = {}
    for i, df_i in enumerate(dfs):
        col_sets[i] = list(df_i.columns)

    reference_cols = None
    for ix, cols in col_sets.items():
        if reference_cols is None:
            reference_cols = set(cols)
        else:
            if set(cols) != reference_cols:
                raise ValueError(f"columns must match to stack dfs, {col_sets}")

    # cast all dtypes to first df
    reference_schema = dfs[0].schema
    sorted_col_exprs = []
    for col in sorted(list(reference_cols)):
        dtype = reference_schema.get(col)
        if dtype is None:
            raise ValueError(f"'{col}' missing from reference schema?")
        expr = pl.col(col).cast(dtype)
        sorted_col_exprs.append(expr)

    dfs_to_stack = [
        df_i.lazy().select(sorted_col_exprs) for df_i in dfs
    ]
    return pl.concat(items=dfs_to_stack, how="vertical", ).lazy()



from __future__ import annotations
import typing as t
import enum
import types
import typing_extensions as t_ext
import pydantic
import pydantic_core
import inflection

from spice_rack._bases import _special_str
from spice_rack._bases import _base_base, _value_model


__all__ = (
    "DispatchedModelMixin",
    "ClassId",
    "DispatchedClassContainer",
    "DispatchableValueModelBase"
)


class ClassId(_special_str.SpecialStrBase):
    """the unique class id for a member of the dispatched family"""
    ...


ClassIdPathT = t.Tuple[ClassId, ...]


SelfTV = t.TypeVar("SelfTV", bound="DispatchedModelMixin")

_LiteralT: t_ext.TypeAlias = type(t.Literal["xxx"])  # type: ignore


class ClassType(enum.Enum):
    """
    the 'type' of class possible when within a family tree of DispatchedModelMixin
    """
    SUPER_ROOT = "super_root"
    """
    sits above the family tree we are creating. The entire 
    """

    ROOT = "root"
    """the first child of DispatchedModelMixin and potentially further roots."""

    CONCRETE = "concrete"
    """actual concrete type of the dispatched family tree. 
    You cannot inherit from this class it is Final
    """


# this has to be setup exactly as is. Anything overwriting __get_pydantic_core_schema__ will
# break the way we generate the pydantic core schema. I am not sure exactly why, but for now
# leave as is. This current setup does cause issues with pydantic_autodoc though.
class _ClassIdStr(str):
    ...
    # @classmethod
    # def __get_pydantic_core_schema__(
    #         cls, _: t.Any, __: pydantic.GetCoreSchemaHandler
    # ) -> pydantic_core.CoreSchema:
    #     return pydantic_core.core_schema.str_schema()


class _PydanticSchemaGenerator(pydantic.GenerateSchema):
    _specified_class_id: t.ClassVar[t.Optional[str]] = None

    def __class_getitem__(cls, item: ClassId) -> t.Type[_PydanticSchemaGenerator]:
        new_cls = types.new_class(f"{item}_schema_gen", bases=(cls, ))
        assert issubclass(new_cls, _PydanticSchemaGenerator), \
            f"{new_cls} not subclass of {_PydanticSchemaGenerator}?"
        new_cls._specified_class_id = str(item)
        return new_cls

    def match_type(self, obj: t.Type[t.Any]) -> pydantic_core.core_schema.CoreSchema:
        if obj is _ClassIdStr:
            return self.gen_class_id_schema()
        else:
            return super().match_type(obj)

    def gen_class_id_schema(self) -> pydantic_core.core_schema.CoreSchema:
        if self._specified_class_id is None:
            return pydantic_core.core_schema.str_schema()
        else:
            return pydantic_core.core_schema.with_default_schema(
                schema=pydantic_core.core_schema.literal_schema(
                    expected=[self._specified_class_id],
                ),
                default=self._specified_class_id
            )
        

_model_config = _base_base.BASE_MODEL_CONFIG
_model_config["schema_generator"] = _PydanticSchemaGenerator


class DispatchedModelMixin(_base_base.PydanticBase):
    """
    the entrypoint for creating dispatchable family trees using pydantic's internal
    engine for schema generation. Be careful not to inherit from two different
    dispatchable family trees at once. It will cause weird behavior and there
    are no checks to protect against this.

    """
    model_config = _model_config

    _class_type: t.ClassVar[ClassType] = ClassType.SUPER_ROOT
    _class_id: t.ClassVar[ClassId]
    _class_id_path: t.ClassVar[ClassIdPathT] = ()

    class_id: _ClassIdStr = pydantic.Field(
        description="this is the field we dispatch the different sub-classes on. "
                    "Concrete classes will constrain this to be only one specific value.",
        default=None,
    )

    @classmethod
    def _get_direct_parent(cls) -> t.Type[DispatchedModelMixin]:
        """iterate over MRO, find most recent DispatchedModelMixin subclass"""
        for _ancestor_cls in cls.mro()[1:]:
            if issubclass(_ancestor_cls, DispatchedModelMixin):
                return _ancestor_cls
        raise ValueError(
            "no subclass of 'DispatchedModelMixin' found?"
        )

    @classmethod
    def _infer_class_id(cls) -> ClassId:
        return ClassId(inflection.underscore(cls.__name__))

    @classmethod
    def _infer_class_type(cls) -> ClassType:
        parent_class = cls._get_direct_parent()
        parent_class_type = parent_class.get_class_type()

        # if super root parent, we assume root
        if parent_class_type == ClassType.SUPER_ROOT:
            class_type_actual = ClassType.ROOT

        # if root parent, we assume concrete
        elif parent_class_type == ClassType.ROOT:
            if cls.__name__.startswith("Abstract"):
                class_type_actual = ClassType.ROOT
            else:
                class_type_actual = ClassType.CONCRETE

        # if concrete parent, we raise ValueError
        elif parent_class_type == ClassType.CONCRETE:
            raise ValueError(
                f"'{cls.__name__}' direct parent is '{parent_class.__name__}' which is a"
                f" concrete class, so we shouldn't be inheriting from it."
            )
        else:
            raise ValueError(f"'{parent_class_type}' unexpected class type value")
        return class_type_actual

    def __init_subclass__(
            cls,
            class_type: t.Optional[t.Union[str, ClassType]] = None,
            class_id: t.Optional[t.Union[str, ClassType]] = None,
    ) -> None:
        """
        control how this subclass is set up.

        Args:
            class_type: specifies the class_id for the class. If specified, we'll use it
                otherwise it'll be constructed from the cls.__name__.
            class_id: If specified, we'll ensure it is valid based on the parent class' ClassType.
                See table below for more info

        Returns: None


        .. list-table:: Class Type specific behavior
           :widths: 100 25 25 25
           :header-rows: 1

           * - Class Type
             - Behavior
             - Possible Parents
             - Possible Children
           * - SUPER_ROOT
             - sits above the family tree. Subclass DispatchedModelMixin
                of the dispatching for entire trees.
             - SUPER_ROOT only
             - SUPER_ROOT or ROOT
           * - ROOT
             - means a class is not concrete, thus it cannot be instantiated directly
             - SUPER_ROOT or ROOT
             - ROOT or CONCRETE
           * - CONCRETE
             - means this is a "leaf" in the family tree. This class can be
                instantiated, and it should not be subclassed further.
             - ROOT only
             - None
        """

        class_id_actual: ClassId
        if class_id:
            class_id_actual = ClassId(class_id)
        else:
            class_id_actual = cls._infer_class_id()

        class_type_actual: ClassType
        if class_type:
            if isinstance(class_type, str):
                class_type_actual = ClassType(class_type.lower())
            else:
                class_type_actual = class_type
        else:
            class_type_actual = cls._infer_class_type()

        cls._class_id = class_id_actual
        cls._class_type = class_type_actual

        # if class_type_actual is CONCRETE, we adjust the schema gen stuff
        if class_type_actual == ClassType.CONCRETE:
            cls.model_config["schema_generator"] = _PydanticSchemaGenerator[cls._class_id]

        # if class_type is SUPER_ROOT, we initialize the class id path
        if class_type_actual == ClassType.SUPER_ROOT:
            cls._class_id_path = ()

        # if it is anything else, we set it using parent class' path
        else:
            parent_cls = cls._get_direct_parent()
            cls._class_id_path = parent_cls._class_id_path + (cls._class_id, )

    @classmethod
    def get_class_type(cls) -> ClassType:
        return cls._class_type

    @classmethod
    def get_class_id(cls) -> ClassId:
        return cls._class_id

    @classmethod
    def get_class_id_path(cls) -> ClassIdPathT:
        return cls._class_id_path

    @classmethod
    def is_concrete(cls) -> bool:
        return cls._class_type == ClassType.CONCRETE

    @classmethod
    def iter_concrete_subclasses(cls: t.Type[SelfTV]) -> t.Iterator[t.Type[SelfTV]]:
        if cls._class_type == ClassType.SUPER_ROOT:
            raise ValueError(
                f"don't call iter_concrete_subclasses on super root classes, "
                f"which '{cls.__name__}' is "
            )
        if cls.is_concrete():
            yield cls
        else:
            for sub_cls in cls.__subclasses__():
                for sub_sub_cls in sub_cls.iter_concrete_subclasses():
                    yield sub_sub_cls

    @pydantic.model_validator(mode="before")
    def _ensure_class_id(cls, data: t.Any) -> t.Any:
        if isinstance(data, dict):
            class_id_found = data.get("class_id")
            if not class_id_found:
                data["class_id"] = str(cls._class_id)
        return data

    @classmethod
    def model_validate(
        cls: t.Type[SelfTV],
        obj: t.Any,
        *args,
        **kwargs,
    ) -> SelfTV:
        if cls._class_type != ClassType.CONCRETE:
            raise ValueError(
                f"'{cls.__name__}' is not concrete, so we do not want"
                f" to initialize instances of this class"
            )
        else:
            return t.cast(SelfTV, super().model_validate(obj, *args, **kwargs))

    @classmethod
    def build_dispatched_ann(cls: t.Type[SelfTV]) -> t.Type[SelfTV]:
        """
        build a type annotation that is equivalent to Union[all concrete subclasses of this class]
        using class_id as a discriminator. This can be used as an annotation for a field in
        another pydantic class.
        """
        options = list(cls.iter_concrete_subclasses())

        if len(options) == 0:
            raise ValueError(
                f"'{cls.__name__}' has no concrete options"
            )
        if len(options) > 1:
            union_t = t.Union[tuple(options)]  # type: ignore
            annotated_t = t.Annotated[
                union_t,  # type: ignore
                pydantic.Discriminator("class_id")
            ]
            return annotated_t  # type: ignore
        else:
            return options[0]

    @classmethod
    def build_dispatcher_type_adapter(cls: t.Type[SelfTV]) -> pydantic.TypeAdapter[SelfTV]:
        """
        use the type annotation created by `build_dispatched_ann` in a pydantic TypeAdapter
        that allows us to get the "discriminated union" parsing functionality outside a
        pydantic class context. see pydantic type adapter docs
        """
        return pydantic.TypeAdapter(cls.build_dispatched_ann())

    @classmethod
    def get_all_concrete_subclass_ids(cls) -> t.List[ClassId]:
        res = []
        for concrete_sub_c in cls.iter_concrete_subclasses():
            res.append(
                concrete_sub_c.get_class_id()
            )
        return res


DispatchedClsTV = t.TypeVar("DispatchedClsTV", bound=DispatchedModelMixin)


class DispatchedClassContainer(pydantic.RootModel[DispatchedClsTV], t.Generic[DispatchedClsTV]):
    """
    convenience wrapper allowing you to specify a dispatched family root class as the class
    parameter, and it'll build the correct discriminated union type under the hood for you.
    """
    root: DispatchedClsTV = pydantic.Field(
        description="the underlying concrete class instance",
    )

    def __class_getitem__(
            cls,
            item: t.Type[DispatchedModelMixin]  # type: ignore
    ) -> t.Type[DispatchedClsTV]:
        if not issubclass(item, DispatchedModelMixin):
            raise ValueError(f"item must be subclass of DispatchedModelMixin, but {item} is not")
        return super().__class_getitem__(
            item.build_dispatched_ann()
        )  # type: ignore


class DispatchableValueModelBase(DispatchedModelMixin, class_type=ClassType.SUPER_ROOT):
    """
    convenience class that combines dispatchable model mixin and value model,
    which is a common use-case for dispatchable models. This has the value model config
    and the `DispatchedModelMixin` functionality.
    """
    model_config = _value_model.VALUE_MODEL_CONFIG



from __future__ import annotations
import typing as t
import dateparser
import datetime as dt

import pydantic
import pydantic_core
from spice_rack import _bases
from spice_rack._ts_service._tz_key import TimeZoneKey
from spice_rack._ts_service._timestamp import Timestamp

__all__ = (
    "StrParserException",
    "StrParser",
    "StrParserStrict",
    "TzValidator"
)


class _StrParserErrorInfo(_bases.exceptions.ErrorInfoBase):
    raw_str: str = pydantic.Field(description="the string we failed to parse into date format")
    strict: bool = pydantic.Field(description="if true, this failed strict mode parsing")
    specified_date_formats: t.List[str] = pydantic.Field(description="the date formats specified ")


class StrParserException(_bases.exceptions.CustomExceptionBase[_StrParserErrorInfo]):
    def __init__(self, raw_str: str, strict: bool, specified_date_formats: t.List[str]):
        error_info = {
            "raw_str": raw_str,
            "strict": strict,
            "specified_date_formats": specified_date_formats
        }

        if strict:
            detail = f"the strict string parser failed to parse '{raw_str}'"
        else:
            detail = f"the flexible string parser failed to parse '{raw_str}'"
        super().__init__(
            detail=detail,
            error_info=error_info,
            verbose=True,
            extra_info=None
        )


class StrParser:
    """
    flexible string parser using dateparser library under the hood. If you provide date_formats it will be as
    fast as a strict parser, but fallback to the slower approach if needed. Differs from StrParserStrict which will
    raise an error not fallback.

    Use this create an annotated Timestamp type within the pydantic validation framework.

    Examples:
        1. Annotate the type to parse the format::

            class Model(pydantic.BaseModel):
                ts: Timestamp
                ts_flexible: t.Annotated[Timestamp, StrParser(date_formats="%m/%d/%Y")]

            date_str = "08/13/1995"
            Model(
                ts=date_str,  # will fail validation
                annotated_ts=date_str  # will be parsed
            )

        2. StrParserStrict will not use dateparser library as a fallback::

            class Model(pydantic.BaseModel):
                ts_strict: t.Annotated[Timestamp, StrictStrParser(date_formats="%m/%d/%Y")]
                ts_flexible: t.Annotated[Timestamp, StrParser(date_formats="%m/%d/%Y")]

            date_str = "08/13/1995"
            Model(
                ts_strict=date_str_good,  # will be parsed
                ts_flexible=date_str  # will be parsed
            )

            date_str = "Aug 13th 1995"
            Model(
                ts_strict=date_str_good,  # will fail validation
                ts_flexible=date_str  # will be parsed
            )
    """
    date_formats: t.List[str]
    """specify these to indicate that the str parser should try these date formats first"""

    def __init__(self, date_formats: t.Optional[t.List[str]] = None):
        self.date_formats = date_formats if date_formats else []

    def _get_func(self) -> t.Callable[[t.Any, ], t.Any]:
        def func(data: t.Any) -> t.Any:
            if isinstance(data, str):
                parsed_maybe = dateparser.parse(data, date_formats=self.date_formats, )
                if parsed_maybe is None:
                    raise StrParserException(
                        raw_str=data,
                        specified_date_formats=self.date_formats,
                        strict=False
                    ).as_pydantic_error()
                else:
                    data = parsed_maybe
            return data
        return func

    def __get_pydantic_core_schema__(
            self,
            __source: t.Any,
            __handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.CoreSchema:
        val_func = self._get_func()

        # todo: add info to the schema
        schema = __handler(__source)
        schema = pydantic_core.core_schema.no_info_before_validator_function(
            function=val_func,
            schema=schema
        )
        return schema


class StrParserStrict:
    """
    This will try the different string formats in order, and choose the first one that works.

    Examples:
        1. Annotate the type to parse the format::

            class Model(pydantic.BaseModel):
                ts: Timestamp
                ts_flexible: t.Annotated[Timestamp, StrParser(date_formats="%m/%d/%Y")]

            date_str = "08/13/1995"
            Model(
                ts=date_str,  # will fail validation
                annotated_ts=date_str  # will be parsed
            )

        2. StrParserStrict will not use dateparser library as a fallback::

            class Model(pydantic.BaseModel):
                ts_strict: t.Annotated[Timestamp, StrictStrParser(date_formats="%m/%d/%Y")]
                ts_flexible: t.Annotated[Timestamp, StrParser(date_formats="%m/%d/%Y")]

            date_str = "08/13/1995"
            Model(
                ts_strict=date_str_good,  # will be parsed
                ts_flexible=date_str  # will be parsed
            )

            date_str = "Aug 13th 1995"
            Model(
                ts_strict=date_str_good,  # will fail validation
                ts_flexible=date_str  # will be parsed
            )
    """
    date_formats: t.List[str]
    """specify these to indicate that the str parser should try these date formats first"""

    def __init__(self, date_formats: t.List[str]):
        if len(date_formats) == 0:
            raise ValueError(
                "You must specify at least one date_format for the StrParserStrict validator"
            )
        self.date_formats = date_formats

    def _get_func(self) -> t.Callable[[t.Any, ], t.Any]:
        def func(data: t.Any) -> t.Any:
            if isinstance(data, str):
                parsed_maybe: t.Optional[dt.datetime] = None
                for format_i in self.date_formats:
                    try:
                        parsed_maybe = dt.datetime.strptime(data, format_i)
                    except ValueError:
                        continue
                    except Exception as e:
                        raise e
                if parsed_maybe is None:
                    raise StrParserException(
                        raw_str=data,
                        specified_date_formats=self.date_formats,
                        strict=True
                    ).as_pydantic_error()
                else:
                    data = parsed_maybe
            return data
        return func

    def __get_pydantic_core_schema__(
            self,
            __source: t.Any,
            __handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.CoreSchema:
        val_func = self._get_func()

        # todo: add info to the schema
        schema = __handler(__source)
        schema = pydantic_core.core_schema.no_info_before_validator_function(
            function=val_func,
            schema=schema
        )
        return schema


class TzValidator:
    """use as an AfterValidator to ensure the timezone is what is expected"""
    def __init__(self, expected_tz: t.Union[str, TimeZoneKey], coerce: bool = True):
        self.expected_tz = TimeZoneKey(expected_tz)
        self.coerce = coerce

    def _get_func(self) -> t.Callable[[t.Any, ], t.Any]:
        def func(ts_inst: Timestamp) -> Timestamp:
            if ts_inst.tz != self.expected_tz:
                if self.coerce:
                    ts_inst = ts_inst.with_tz(self.expected_tz)
                else:
                    raise ValueError(
                        f"expected '{self.expected_tz}' timezone, encountered '{ts_inst.tz}'"
                    )
            return ts_inst
        return func

    def __get_pydantic_core_schema__(
            self,
            __source: t.Any,
            __handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.CoreSchema:
        val_func = self._get_func()

        # todo: add info to the schema
        schema = __handler(__source)
        schema = pydantic_core.core_schema.no_info_after_validator_function(
            function=val_func,
            schema=schema
        )
        return schema



from __future__ import annotations
import pydantic
from google.auth import exceptions as gcp_exceptions  # noqa

from spice_rack import _bases


__all__ = (
    "DefaultAuthNotAvailableException",
)


class _DefaultAuthNotAvailableErrorInfo(_bases.exceptions.ErrorInfoBase):
    gcp_error: _bases.exceptions.WrappedExternalException = pydantic.Field(
        description="error from the gcp auth external lib"
    )


class DefaultAuthNotAvailableException(_bases.exceptions.CustomExceptionBase[_DefaultAuthNotAvailableErrorInfo]):
    def __init__(
            self,
            detail: str,
            gcp_error: gcp_exceptions.GoogleAuthError,
    ):
        error_info = {
            "gcp_error": gcp_error
        }
        super().__init__(
            detail=detail if detail else str(gcp_error),
            error_info=error_info,
            verbose=True,
            extra_info=None
        )



from spice_rack._polars_service._types._dicts import *
from spice_rack._polars_service._types._collected_df import *
from spice_rack._polars_service._types._lazy_df import *
from spice_rack._polars_service._types._maybe_lazy import *



from __future__ import annotations
import typing as t
import zoneinfo
import tzlocal

from spice_rack import _bases


__all__ = (
    "TimeZoneKey",
    "InvalidTimeZoneKeyException"
)


_TZ_CHOICES_SET: t.Set[str] = zoneinfo.available_timezones()
_TZ_CHOICES_LIST: t.List[str] = list(_TZ_CHOICES_SET)


class TimeZoneKey(_bases.special_str.SpecialStrBase):
    """
    a special string representing a timezone.

    see python docs for "zoneinfo" for options:
        https://docs.python.org/3/library/zoneinfo.html
    """

    @classmethod
    def timezone_choices(cls) -> t.Set[str]:
        return _TZ_CHOICES_SET

    @classmethod
    def _format_str_val(cls, root_data: str) -> str:
        # ensure it is a valid zone info
        if root_data not in cls.timezone_choices():
            raise InvalidTimeZoneKeyException(
                invalid_value=root_data,
                options=_TZ_CHOICES_LIST
            )

        else:
            return root_data

    def as_py_zone_info(self) -> zoneinfo.ZoneInfo:
        """get the timezone info in the format python datetime library uses"""
        return zoneinfo.ZoneInfo(str(self))

    @classmethod
    def local(cls) -> TimeZoneKey:
        """gets timezone info for where this process is running"""
        return _LOCAL_TZ_KEY


# set this once on import bc it is pretty slow
_LOCAL_TZ_KEY = TimeZoneKey(tzlocal.get_localzone().key)


class _InvalidTimeZoneKeyErrorInfo(_bases.exceptions.ErrorInfoBase):
    invalid_value: str
    options: list[str]


class InvalidTimeZoneKeyException(_bases.exceptions.CustomExceptionBase[_InvalidTimeZoneKeyErrorInfo]):
    def __init__(
            self,
            invalid_value: str,
            options: list[str],
            extra_info: t.Optional[dict] = None,
            verbose: bool = False
    ):
        error_info = {
            "invalid_value": invalid_value,
            "options": options
        }

        detail = f"'{invalid_value}' is not a valid timezone key."
        super().__init__(
            detail=detail,
            error_info=error_info,
            extra_info=extra_info,
            verbose=verbose
        )



from __future__ import annotations
import typing as t
import pydantic
import stackprinter
import devtools

from spice_rack import _bases

__all__ = (
    "ExtraLogData",
    "TRACEBACK_AUG_STR_ID",
    "ExtraLogDataContainer",
)


TRACEBACK_AUG_STR_ID = "traceback"


class ExtraLogData(_bases.value_model.ValueModelBase):
    """general purpose log augmentation object"""
    key: str = pydantic.Field(
        description="the key we use for the log aug formatted data",
    )
    desc: str = pydantic.Field(description="the description", default="no description")
    data: t.Any = pydantic.Field(
        description="the data for this log aug object, must be json encode-able"
    )

    @pydantic.computed_field
    def data_type(self) -> str:
        """the class name of the data type"""
        return type(self.data).__name__

    def get_serializable_data(self) -> dict:
        return self.json_dict(use_str_fallback=True)

    @classmethod
    def parse_from_raw_data(cls, data: t.Any) -> ExtraLogData:
        # parse raw data of different types

        # if we find formatting logic, use that
        from spice_rack._logging._log_aug._loggable_mixin import LoggableObjMixin
        if isinstance(data, LoggableObjMixin):
            data = data.__format_for_logger__()

        # check types
        if isinstance(data, ExtraLogData):
            data = data

        elif isinstance(data, str):
            if data == TRACEBACK_AUG_STR_ID:
                tb_data = stackprinter.format(show_vals=None, style="plaintext")
                data = {"key": "some_traceback", "data": tb_data.splitlines()}
            else:
                data = {"key": "some_str_data", "data": data}

        # wrap exceptions in our internal wrapper to get better formatting
        elif isinstance(data, Exception):
            from spice_rack import _bases
            wrapped_exc = _bases.exceptions.WrappedExternalException.model_validate(data)
            data = {
                "key": "logged_exception",
                "data": wrapped_exc.json_dict(use_str_fallback=True)
            }

        # we just specify its type and treat it as data
        else:
            data = {"key": "some_data", "data": data}

        return ExtraLogData.model_validate(data)


class ExtraLogDataContainer(pydantic.RootModel[t.List[ExtraLogData]]):
    def get_logger_repr(self) -> str:
        serialized_data = [item.get_serializable_data() for item in self.root]
        return devtools.pformat(serialized_data)



from __future__ import annotations
import typing as t
from abc import abstractmethod

import pydantic_settings

__all__ = (
    "CustomSettingSourceBase",
)


class CustomSettingSourceBase(pydantic_settings.PydanticBaseSettingsSource):
    """base class for a custom settings_base source. matches specified format pydantic wants"""
    @abstractmethod
    def __call__(self) -> dict[str, t.Any]:
        ...



from __future__ import annotations
import typing as t
from typing_extensions import TypeAlias
import warnings

from loguru import _logger as _loguru_module  # noqa
from spice_rack import _bases
from spice_rack._logging import _log_aug, _log_level, _sinks


__all__ = (
    "Logger",
)

LoguruLoggerT: TypeAlias = _loguru_module.Logger

_ExtraLogDataT = t.Optional[list[t.Any]]
_LogLevelOrStrT = t.Union[str, _log_level.LogLevel]


class Logger(_bases.base_base.PydanticBase):
    """
    The main interface. Use the 'setup_logger' classmethod to configure the logger
    and 'get_logger' classmethod to get a Logger instance to use.
    The methods are the standard ones you'd want in a logger.
    """
    _cached_logger: t.ClassVar[t.Optional[Logger]] = None
    """previously set up logger"""

    _logger_wrapper_depth: t.ClassVar[int] = 4
    """ensures we use the part of the stack that actually called the logger"""

    _default_service_name: t.ClassVar[str] = "unknown"
    """the default service name we'll use"""

    service_name: str
    """the name of the service to use"""

    sinks: t.List[_sinks.AnySinkT]
    """
    the sink configs we used for this logger. Merely there for informational purposes as they are used for
    setup prior to initializing this instance
    """

    @classmethod
    def setup_logger(
            cls,
            service_name: str,
            sinks: t.List[_sinks.AnySinkT],
            cache_res: bool = True,
    ) -> Logger:
        """
        Sets up the logger with the sinks and the service name.
        Use this classmethod to initially set up a logger and for all subsequent calls,
        use 'get_logger' class method.

        Args:
            service_name: the name of the service we are setting up the logger for
            sinks: the sink configs we want to use
            cache_res: if True, we cache the Logger inst for later usage. default is True

        Returns:
            Logger: a setup logger inst
        """
        if sinks:
            from spice_rack._logging._setup import configure_logging_for_python_worker
            configure_logging_for_python_worker(sinks=sinks)
        logger_inst = Logger(
            service_name=service_name,
            sinks=sinks,
        )
        if cache_res:
            cls._cached_logger = logger_inst
        return logger_inst

    @classmethod
    def setup_logger_uvicorn(
            cls,
            service_name: str,
            sinks: t.List[_sinks.AnySinkT],
            cache_res: bool = True,
    ) -> Logger:
        """
        Sets up the logger with the sinks and the service name, with special logic for when we're using it within
        an uvicorn server context

        Args:
            service_name: the name of the service we are setting up the logger for
            sinks: the sink configs we want to use
            cache_res: if True, we cache the Logger inst for later usage. default is True

        Returns:
            Logger: a setup logger inst
        """
        if sinks:
            from spice_rack._logging._setup import configure_logging_for_uvicorn_server
            configure_logging_for_uvicorn_server(sinks=sinks)
        logger_inst = Logger(
            service_name=service_name,
            sinks=sinks,
        )
        if cache_res:
            cls._cached_logger = logger_inst
        return logger_inst

    @classmethod
    def get_logger(cls, service_name: t.Optional[str] = None) -> Logger:
        """
        get a previously set up logger. If nothing specified, we'll create a logger with no sink configs and raise a
        warning about it.

        Args:
            service_name: if specified, we'll change the service name to this value, otherwise it'll be unchanged

        Returns: a Logger instance
        """
        logger_inst: Logger

        if cls._cached_logger is None:
            warnings.warn(
                "no logger instance found, we setup a basic logger to use. "
                "Make sure you didn't forget to configure your logger"
            )
            logger_inst = cls.setup_logger(
                service_name=service_name if service_name else cls._default_service_name,
                sinks=[],
                cache_res=True
            )
        else:
            logger_inst = cls._cached_logger

        if service_name:
            logger_inst.service_name = service_name
        return logger_inst

    def _post_init_setup(self) -> None:
        self.loguru_logger.configure(extra={"service_name": self.service_name, "extra_data": ""})

    @property
    def loguru_logger(self) -> LoguruLoggerT:
        """the underlying loguru logger instance"""
        from loguru import logger as loguru_logger
        return loguru_logger

    @staticmethod
    def _format_log_extra_data(
            extra_data: _ExtraLogDataT
    ) -> _log_aug.ExtraLogDataContainer:
        res: list[_log_aug.ExtraLogData] = []
        if extra_data:
            for data_i in extra_data:
                formatted_aug = _log_aug.ExtraLogData.parse_from_raw_data(data_i)
                res.append(formatted_aug)
        return _log_aug.ExtraLogDataContainer.model_validate(res)

    def _get_depth(self, depth_adj: t.Optional[int]) -> int:
        """gets the adjusted depth"""
        depth = self._logger_wrapper_depth
        if depth_adj:
            depth = depth + depth_adj

        if depth < 0:
            raise ValueError(
                f"depth can't be less than 0, encountered {depth}"
            )
        return depth

    def _log(
            self,
            level: _log_level.LogLevel,
            msg: str,
            extra_data: _ExtraLogDataT,
            depth: t.Optional[int] = None,
    ) -> None:
        # catch errors and logs them, without blowing
        try:
            extra_data = self._format_log_extra_data(extra_data).get_logger_repr()

        except Exception as e:
            self._log(
                level=_log_level.LogLevel.model_validate("error"),
                msg="failed to format the specified extra data",
                extra_data=[str(e)],
                # manually specify depth here to be 0 bc this is an error within this logger
                depth=-1*self._logger_wrapper_depth
            )
            return

        try:
            loguru_level: str = str(level).upper()
            loguru_logger_inst = self.loguru_logger.opt(depth=self._get_depth(depth))
            loguru_logger_inst.log(
                loguru_level, msg,
                service_name=self.service_name,
                extra_data=extra_data
            )

        except Exception as e:
            self._log(
                level=_log_level.LogLevel.model_validate("error"),
                msg="error encountered when calling the loguru logger",
                extra_data=[str(e)],

                # manually specify depth here to be 0 bc this is an error within this logger
                depth=-1*self._logger_wrapper_depth
            )
            return

    def log(
            self,
            level: _log_level.LogLevel,
            msg: str,
            extra_data: _ExtraLogDataT = None,
            depth: t.Optional[int] = None
    ) -> None:
        return self._log(
            level=_log_level.LogLevel.model_validate(level),
            msg=msg,
            extra_data=extra_data,
            depth=depth
        )

    def debug(
            self,
            msg: str,
            extra_data: _ExtraLogDataT = None,
            depth: t.Optional[int] = None
    ):
        self.log(
            level=_log_level.LogLevel("debug"),
            msg=msg,
            extra_data=extra_data,
            depth=depth
        )

    def info(
            self,
            msg: str,
            extra_data: _ExtraLogDataT = None,
            depth: t.Optional[int] = None
    ):
        self.log(
            level=_log_level.LogLevel("info"),
            msg=msg,
            extra_data=extra_data,
            depth=depth
        )

    def error(
            self,
            msg: str,
            extra_data: _ExtraLogDataT = None,
            depth: t.Optional[int] = None
    ):
        self.log(
            level=_log_level.LogLevel("error"),
            msg=msg,
            extra_data=extra_data,
            depth=depth
        )

    def critical(
            self,
            msg: str,
            extra_data: _ExtraLogDataT = None,
            depth: t.Optional[int] = None
    ):
        self.log(
            level=_log_level.LogLevel("critical"),
            msg=msg,
            extra_data=extra_data,
            depth=depth
        )

    def exception(
            self,
            msg: str,
            extra_data: _ExtraLogDataT = None,
            depth: t.Optional[int] = None
    ):
        # catch errors and logs them, without blowing
        try:
            extra_data = self._format_log_extra_data(extra_data).get_logger_repr()

        except Exception as e:
            self._log(
                level=_log_level.LogLevel.model_validate("error"),
                msg="failed to format the specified log augmentations",
                extra_data=[str(e)],
                # manually specify depth here to be 0 bc this is an error within this logger
                depth=-1*self._logger_wrapper_depth
            )
            return

        try:
            loguru_level: str = str(_log_level.LogLevel("error")).upper()
            loguru_logger_inst = self.loguru_logger.opt(depth=self._get_depth(depth))
            loguru_logger_inst.log(
                loguru_level, msg,
                service_name=self.service_name,
                extra_data=extra_data
            )

        except Exception as e:
            self._log(
                level=_log_level.LogLevel.model_validate("error"),
                msg="error encountered when calling the loguru logger",
                extra_data=[str(e)],

                # manually specify depth here to be 0 bc this is an error within this logger
                depth=-1*self._logger_wrapper_depth
            )
            return



from __future__ import annotations
import typing as t
from pydantic import Field, validate_call
from fsspec.implementations import sftp
from paramiko.sftp_file import SFTPFile as ParamikoSftpFile
from paramiko.sftp_client import SFTPClient as ParamikoSftpClient

from spice_rack._fs_ops import _path_strs
from spice_rack._fs_ops._file_systems import _base


__all__ = (
    "SftpFileSystem",
)


@t.final
class SftpFileSystem(_base.AbstractFileSystem, class_id="sftp"):
    """wrapper for a paramiko-based sftp file system"""

    host: str = Field(description="the host url of this server")
    username: t.Optional[str] = Field(description="the username to use", default=None)
    password: t.Optional[str] = Field(description="the password to use", default=None)
    port: t.Optional[int] = Field(description="the port to use", default=None)

    @classmethod
    def get_fs_specific_prefix(cls) -> str:
        return "/"

    def build_fsspec_file_system(self) -> sftp.SFTPFileSystem:
        kwargs = {
            "host": self.host
        }
        if self.username:
            kwargs["username"] = self.username

        if self.password:
            kwargs["password"] = self.password

        if self.port:
            kwargs["port"] = self.port

        fs = sftp.SFTPFileSystem(
            **kwargs,

            # missing_host_key_policy=AutoAddPolicy()

        )
        # hacky thing to make sure it isn't closed
        if fs.ftp.sock.closed():
            fs._connect()  # noqa

        return fs

    def get_home_dir(self) -> _path_strs.AbsoluteDirPathStr:
        # todo: revisit this
        return _path_strs.AbsoluteDirPathStr("/")

    def download_file_locally(
            self,
            path: _path_strs.AbsoluteFilePathStr,
            dest_dir: _path_strs.AbsoluteDirPathStr
    ) -> _path_strs.AbsoluteFilePathStr:
        from spice_rack._fs_ops._file_systems import _local
        local_fs = _local.LocalFileSystem()
        local_fs.make_dir(dest_dir, if_exists="return", create_parents=True)

        local_path = dest_dir.joinpath(
            _path_strs.RelFilePathStr(path.get_name(include_suffixes=True))
        )

        fsspec_inst = self.build_fsspec_file_system()
        paramiko_client: ParamikoSftpClient = fsspec_inst.ftp

        _paramiko_sftp_get(
            sftp_client=paramiko_client,
            sftp_file=self.contextualize_abs_path(path),
            local_file=str(local_path),
        )
        return local_path

    @validate_call
    def download_dir_locally(
            self,
            __source_dir: _path_strs.AbsoluteDirPathStr,
            __local_dest_dir: _path_strs.AbsoluteDirPathStr
    ) -> _path_strs.AbsoluteDirPathStr:
        from spice_rack._fs_ops._file_systems import _local
        local_fs = _local.LocalFileSystem()
        local_fs.make_dir(__source_dir, if_exists="return", create_parents=True)

        local_path = __local_dest_dir.joinpath(
            _path_strs.RelDirPathStr(__source_dir.get_name())
        )

        for source_abs_file_path in self.iter_dir_contents_files_only(
                __source_dir, recursive=True
        ):
            rel_file_path = _path_strs.RelFilePathStr(
                str(source_abs_file_path).replace(str(__source_dir), "")
            )
            dest_abs_file_path = local_path.joinpath(rel_file_path)
            self.download_file_locally(
                path=source_abs_file_path, dest_dir=dest_abs_file_path.get_parent()
            )
        return local_path


# copied from here: https://github.com/paramiko/paramiko/issues/151#issuecomment-1076144423
# read through issue for full context, but tldr: paramiko fails on large files


def _paramiko_sftp_get(
        sftp_client: ParamikoSftpClient,
        sftp_file: str,
        local_file: str,
        callback: t.Optional[t.Callable[[int, int], None]] = None,
        max_request_size: int = 2 ** 28) -> int:
    """
    A copy of paramiko's sftp.get() function that allows for sequential download
    of large chunks.
    This is a work-around for https://github.com/paramiko/paramiko/issues/151.
    The issue does not occur when prefetch=False (i.e. sequential download) indicating
    that there seems to be an error with the parallel approach. However, the sequential
    version in paramiko does not allow customizable request size, and instead hardcodes a
    small value that is known to work with many SFTP implementations.
    With the possibility of large chunks, the sequential download's RTT overhead becomes
    less of a pain and a viable alternative.
    :param sftp_client: Paramiko's SFTPClient.
    :param sftp_file: The remote file in sftp.
    :param local_file: The local file.
    :param callback: A function that is invoked on every chunk.
    :param max_request_size: The max request size, defaults to 2**20.
    :return: The size of the file in bytes.
    """

    with open(local_file, "wb") as local_handle:
        file_size = sftp_client.stat(sftp_file).st_size
        assert file_size is not None

        with sftp_client.open(sftp_file, "rb") as remove_handle:
            _paramiko_transfer_with_callback(
                remove_handle,
                local_handle,
                file_size,
                callback,
                max_request_size
            )

    return file_size


def _paramiko_transfer_with_callback(
        reader: ParamikoSftpFile,
        writer: "BinaryIO",  # noqa 821 -- not sure where this comes from
        file_size: int,
        callback: t.Optional[t.Callable[[int, int], None]],
        max_request_size: int):
    """
    A copy of paramiko's sftp_client._transfer_with_callback with max_request_size support.
    :param reader: The reader file handle.
    :param writer: The writer file handle.
    :param file_size: The size of the file to be downloaded.
    :param callback: A function that is invoked on every chunk.
    :param max_request_size: The max request size, defaults to 2**20.
    """
    size = 0
    reader.MAX_REQUEST_SIZE = max_request_size

    while True:
        remaining = file_size - size
        chunk = min(max_request_size, remaining)

        data = reader.read(chunk)
        writer.write(data)
        size += len(data)

        if len(data) == 0:
            break

        if callback is not None:
            callback(size, file_size)

    assert size == file_size



from __future__ import annotations
import typing as t
from typing_extensions import TypeAlias
from pydantic import Field

from spice_rack import _bases

if t.TYPE_CHECKING:
    from spice_rack._frozen_registry._base import FrozenRegistryBase

__all__ = (
    "DuplicateKeysInRegistryException",
)


_DupeKeyCountsT: TypeAlias = dict[str, int]


class _DuplicateKeysInRegistryErrorInfo(_bases.exceptions.ErrorInfoBase):
    """duplicate keys found in the registry"""
    registry_type: str = Field(description="the name of the registry class that made this error")
    duplicates_with_counts: _DupeKeyCountsT = Field(
        description="a dict of keys that are duplicate and the number of times they appear"
    )


class DuplicateKeysInRegistryException(
    _bases.exceptions.CustomExceptionBase[_DuplicateKeysInRegistryErrorInfo]
):
    """duplicate keys found in the registry"""
    def __init__(
            self,
            registry_type: t.Type[FrozenRegistryBase],
            duplicate_keys: _DupeKeyCountsT,
            verbose: bool = True,
            extra_info: t.Optional[dict] = None
    ):
        detail = f"{len(duplicate_keys)} found in the '{registry_type.__name__}' registry"
        error_info = {
            "registry_type": registry_type.__name__,
            "duplicates_with_counts": duplicate_keys,
        }
        super().__init__(
            detail=detail,
            error_info=error_info,
            verbose=verbose,
            extra_info=extra_info
        )



from __future__ import annotations
import json
import typing as t
import polars as pl

if t.TYPE_CHECKING:
    from spice_rack._polars_service import _types


__all__ = (
    "get_json_safe_row_dicts",
    "get_json_safe_column_dict",
    "get_json_safe_row_dicts_lazy_peek",
    "get_json_safe_column_dict_lazy_peek"
)


def get_json_safe_row_dicts(
        df: _types.PolarsMaybeLazyDfT,
) -> _types.RowDictsT:
    """
    dump the polars LazyFrame or DataFrame into json-encodeable list of dictionaries
    Args:
        df: LazyFrame or DataFrame instance

    Returns:
        a list of dicts
    """
    collected_df: pl.DataFrame
    if isinstance(df, pl.LazyFrame):
        collected_df = df.collect()  # noqa -- idk what p is
    else:
        collected_df = df

    df_json = collected_df.write_json(file=None, row_oriented=True)
    return json.loads(df_json)


def get_json_safe_column_dict(
        df: _types.PolarsMaybeLazyDfT
) -> _types.ColumnarDictT:
    """
    dump the polars LazyFrame or DataFrame into json-encodeable dict with keys being column names and values
    being lists containing the data for the given column.

    Args:
        df: LazyFrame or DataFrame instance

    Returns:
        a dict of lists
    """
    collected_df: pl.DataFrame
    if isinstance(df, pl.LazyFrame):
        collected_df = df.collect()  # noqa -- idk what p is
    else:
        collected_df = df

    df_json = collected_df.write_json(file=None, row_oriented=False)
    return json.loads(df_json)


def get_json_safe_row_dicts_lazy_peek(
        lazy_df: _types.PolarsLazyDfT, num_rows: int = 3
) -> _types.RowDictsT:
    """
    dump the first n rows of a polars LazyFrame into a json-encodeable list of dictionaries,
    without having to collect the entire LazyFrame.

    Args:
        lazy_df: LazyFrame instance
        num_rows: the number of rows to collect

    Returns:
        a list of dicts
    """
    sample_df = lazy_df.fetch(n_rows=num_rows)
    return get_json_safe_row_dicts(sample_df)


def get_json_safe_column_dict_lazy_peek(
        lazy_df: _types.PolarsMaybeLazyDfT, num_rows: int = 3
) -> _types.ColumnarDictT:
    """
    dump the first n rows of a polars LazyFrame into a json-encodeable dict with keys being column names and values
    being lists containing the data for the given column.

    Args:
        lazy_df: LazyFrame instance
        num_rows: the number of rows to collect

    Returns:
        a dict of lists
    """
    sample_df = lazy_df.fetch(n_rows=num_rows)
    return get_json_safe_column_dict(sample_df)



from __future__ import annotations
import typing as t
from pydantic import Field

from spice_rack._bases._value_model import ValueModelBase


__all__ = (
    "ErrorInfoBase",
    "ErrorPayload"
)


class ErrorInfoBase(ValueModelBase):
    """base object that holds information about a specific error."""
    extra: dict = Field(
        description="dict field where you can add extra info. "
                    "Can include any data types but everything must be json serializable",
        default_factory=dict
    )

    def get_json_data(self, **pydantic_kwargs) -> dict:
        return self.json_dict(use_str_fallback=True, **pydantic_kwargs)


ErrorInfoTV = t.TypeVar("ErrorInfoTV", bound=ErrorInfoBase)


class ErrorPayload(ValueModelBase, t.Generic[ErrorInfoTV]):
    """
    Generic payload class holding the error detail string, and the info object.
    The ErrorInfoTV type var dictates the unique
    schema information about the concrete class
    """
    _custom_schema_title: t.ClassVar[t.Optional[str]] = None
    detail: str = Field(description="the top line error")

    error_type: str = Field(
        description="the name of the error class this payload is connected to"
    )
    info: ErrorInfoTV = Field(
        description="object containing more detailed information about the error"
    )



from __future__ import annotations
import typing as t
import pydantic
from google.auth import credentials, default as gcp_default, exceptions as gcp_exceptions  # noqa

from spice_rack._gcp_auth._auth_strategies import _base
from spice_rack._gcp_auth import _exceptions


__all__ = (
    "DefaultAuthStrategy",
)


def _get_gcp_default_creds() -> credentials.Credentials:
    """fixes type annotation for the function provided by gcp"""
    credentials_inst, _ = gcp_default()
    return credentials_inst


class DefaultAuthStrategy(_base.AbstractGcpAuthStrategy):
    """
    just ride with whatever defaults we have, works if we are in a vm and things like that
    see `docs <https://googleapis.dev/python/google-api-core/latest/auth.html#authentication>`_

    If nothing found we'll use anonymous.
    """
    on_no_default:  t.Literal["raise", "use_anon"] = pydantic.Field(
        description="What to do if GCP's default credential builder fails to find anything"
                    " in the environment. Choices are: 'raise', and 'use_anon'",
        default="use_anon"
    )

    def default_creds_found(self) -> bool:
        try:
            credentials_inst = _get_gcp_default_creds()
            if credentials_inst.token_state == credentials.TokenState.STALE:
                return False

        except gcp_exceptions.DefaultCredentialsError:
            return False

        except Exception as e:
            raise e

    def build_gcp_creds(
            self
    ) -> credentials.Credentials:
        """
        builds the GCP credentials object.

        Returns: The GCP Credentials instance

        Raises:
            DefaultAuthNotAvailableException: if we fail to find valid creds in the environment and the
                "on_no_default" is set to "raise"
        """
        # see:
        #   https://googleapis.dev/python/google-api-core/latest/auth.html#user-accounts-3-legged-oauth-2-0-with-a-refresh-token

        try:
            credentials_inst = _get_gcp_default_creds()
            if credentials_inst.token_state == credentials.TokenState.STALE:
                raise gcp_exceptions.RefreshError()

        except gcp_exceptions.GoogleAuthError as e:
            if self.on_no_default == "raise":
                detail: str

                if isinstance(e, gcp_exceptions.DefaultCredentialsError):
                    detail = "unable to get default gcp credentials from the environment"
                elif isinstance(e, gcp_exceptions.RefreshError):
                    detail = "default gcp credentials found, but they are stale. refresh them using gcloud"
                else:
                    detail = "encountered unexpected google auth error"

                raise _exceptions.DefaultAuthNotAvailableException(
                        detail=detail,
                        gcp_error=e
                    )

            elif self.on_no_default == "use_anon":
                credentials_inst = credentials.AnonymousCredentials()

            else:
                raise ValueError(
                    f"unexpected value for 'on_no_default' param: '{self.on_no_default}'"
                )

        return credentials_inst



from __future__ import annotations
from typing import Any, Type, TypeVar


__all__ = (
    "check_type",
    "check_subclass",
    "is_subclass_strict",
)


_TypeClsTV = TypeVar("_TypeClsTV", )


def check_type(obj: Any, cls: Type[_TypeClsTV]) -> _TypeClsTV:
    if isinstance(obj, cls):
        return obj
    else:
        raise TypeError(
            f"expected obj to be type '{cls.__name__}' but encountered {obj.__class__.__name__}"
        )


def check_subclass(
        cls_in_question: Type[Any],
        cls_to_enforce: Type[_TypeClsTV]
) -> Type[_TypeClsTV]:
    if not isinstance(cls_in_question, type):
        raise TypeError(
            f"cls_in_question param is not even a class. it must be. "
            f"encountered: {cls_in_question} "
            f"which is type: '{cls_in_question.__class__.__name__}'"
        )
    if issubclass(cls_in_question, cls_to_enforce):
        return cls_in_question
    else:
        raise TypeError(
            f"expected obj to be subclass of '{cls_to_enforce.__name__}' "
            f"but encountered {cls_in_question.__class__.__name__} which isn't"
        )


def is_subclass_strict(child_cls: type, parent_cls: type) -> bool:
    """
    checks if the child class inherits from the parent class
    but is not the parent class exactly
    """
    if child_cls == parent_cls:
        return False
    return issubclass(child_cls, parent_cls)



from __future__ import annotations
import typing as t
import pydantic
import pydantic_core

from spice_rack._fs_ops import _file_info, _exceptions

if t.TYPE_CHECKING:
    from spice_rack._fs_ops import _fs_models, _path_strs


__all__ = (
    "ExistsConstraint",
    "FileExtConstraint",
)


# these cannot be pydantic models themselves


class ExistsConstraint:
    """ensures the fs obj path exists"""

    @staticmethod
    def _get_func() -> pydantic_core.core_schema.NoInfoValidatorFunction:
        def _func(fs_obj: _fs_models.FileOrDirPathT) -> _fs_models.FileOrDirPathT:
            try:
                fs_obj.ensure_exists()

            except _exceptions.NonExistentPathException as e:
                raise e.as_pydantic_error()

            except Exception as e:
                raise e

            return fs_obj
        return _func

    def __get_pydantic_core_schema__(
            self,
            __source: t.Any,
            __handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.CoreSchema:
        val_func = self._get_func()

        # todo: add info to the schema
        schema = __handler(__source)
        schema = pydantic_core.core_schema.no_info_after_validator_function(
            function=val_func,
            schema=schema
        )
        return schema


class _FileExtValFunc:
    """used in file extension constraints"""
    def __init__(self, options: t.List[_file_info.FileExt]):
        self.options = options

    def __call__(self, __obj: t.Union[_fs_models.FilePath, _path_strs.RelOrAbsFilePathT]) -> _fs_models.FilePath:
        try:
            __obj.ensure_correct_file_ext(choices=self.options)

        except _exceptions.InvalidFileExtensionException as e:
            raise e.as_pydantic_error()

        except Exception as e:
            raise e

        return __obj


_OptionsTypeAdapter = pydantic.TypeAdapter(t.List[_file_info.FileExt])


class _FileExtConstraintBase:
    """base class for a file extension constraint class"""
    _file_ext_options_md_key: t.ClassVar[str] = "valid_file_extensions"
    """key we use in the pydantic schema metadata for the file extension options"""
    def __init__(self, __options: t.List[str]):
        # ensure options is at least one
        self._options: t.List[_file_info.FileExt] = _OptionsTypeAdapter.validate_python(__options)

    def _update_pydantic_schema_metadata(
            self,
            pydantic_schema: pydantic_core.CoreSchema,
    ) -> pydantic_core.CoreSchema:
        _existing_schema_md = pydantic_schema.get("metadata")
        schema_md = _existing_schema_md if _existing_schema_md else {}

        # if not a dict we do nothing bc I'm not sure what's going on
        if isinstance(schema_md, dict):
            schema_md["valid_file_extensions"] = self._options
            pydantic_schema["metadata"] = schema_md
        return pydantic_schema

    def __get_pydantic_core_schema__(
            self,
            __source: t.Any,
            __handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.CoreSchema:
        val_func = _FileExtValFunc(options=self._options)
        schema = __handler(__source)
        schema = pydantic_core.core_schema.no_info_after_validator_function(
            function=val_func,
            schema=schema
        )
        schema = self._update_pydantic_schema_metadata(schema)
        return schema

    def __get_pydantic_json_schema__(
            self,
            core_schema: pydantic_core.CoreSchema,
            handler: pydantic.GetJsonSchemaHandler
    ) -> pydantic.json_schema.JsonSchemaValue:
        json_schema = handler(core_schema)
        json_schema = handler.resolve_ref_schema(json_schema)
        current_desc = json_schema.get("description", "")

        description_add_on: str
        if len(self._options) > 1:
            file_ext_options_strs = [f"'{str(option)}'" for option in self._options]
            file_ext_options_str = ", ".join(file_ext_options_strs)
            description_add_on = f"the file must have one of the following extensions: [{file_ext_options_str}]"
        else:
            description_add_on = f"the file must have the '{str(self._options[0])}' file extension"

        updated_desc = f"{current_desc}\n{description_add_on}"
        json_schema["description"] = updated_desc
        return json_schema


class FileExtConstraint(_FileExtConstraintBase):
    """generic file ext constraint, specify the options yourself"""
    def __init__(self, __options: t.Union[str, t.Iterable[str]]):
        if not __options:
            raise ValueError(
                "must specify a list of valid file extensions"
            )

        choices: t.List[_file_info.FileExt]
        if isinstance(__options, str):
            choices = [_file_info.FileExt(__options)]
        elif isinstance(__options, t.Iterable):
            choices = [_file_info.FileExt(c) for c in __options]
        else:
            raise ValueError(
                f"must specify valid file extensions as an iterable of strings or a str. received '{__options}'"
            )

        super().__init__(choices)



from __future__ import annotations

from spice_rack._bases._base_base import PydanticBase


__all__ = ("PydanticMixinBase", )


class PydanticMixinBase(PydanticBase):
    """
    Base for a class intended to be used purely as a mixin to augment another pydantic model.
    Primarily this is meant to add fields to another model.
    No special behavior is implemented, and this inherits from pydantic's BaseModel
    and behaves as a regular pydantic class. We will circle back and add checks, but for
    now the extra layer of inheritance can at least serve as a visual call out about a given
    mixin class.
    """
    ...



from __future__ import annotations
import typing as t
from pathlib import Path


__all__ = (
    "file_validator",
    "dir_validator",
    "rel_validator",
    "abs_validator"
)


def file_validator(
        raw_str: str
) -> t.List[str]:
    """checks path makes sense as a relative or absolute file"""
    if raw_str.endswith("/"):
        return ["a file path str cannot end in '/'"]
    else:
        return []


def dir_validator(raw_str: str, strict: bool = False) -> t.List[str]:
    if raw_str.endswith("/"):
        return []

    issue: t.Optional[str] = None
    if Path(raw_str).suffix:
        issue = "this path has a file extension suffix, so it cannot be a directory path"
    else:
        if strict:
            issue = "a dir path must end in '/'"

    if issue is not None:
        return [issue]
    else:
        return []


def rel_validator(raw_str: str) -> t.List[str]:
    if raw_str.startswith("/"):
        return [
            "a relative path cannot start with '/'"
        ]
    else:
        return []


def abs_validator(raw_str: str) -> t.List[str]:
    if not raw_str.startswith("/"):
        return [
            "an absolute path must start with '/'"
        ]
    else:
        return []



from __future__ import annotations
import typing as t
import pydantic
import polars as pl

from spice_rack._polars_service._types import _collected_df, _lazy_df


__all__ = (
    "PolarsMaybeLazyDfT",
    "PolarsMaybeLazyDfTypeAdapter",
    "PolarsDfTV"
)


def _callable_discrim(_obj: t.Any) -> t.Literal["collected", "lazy"]:
    if isinstance(_obj, pl.LazyFrame):
        return "lazy"
    else:
        return "collected"


PolarsMaybeLazyDfT = t.Annotated[
    t.Union[
        t.Annotated[_collected_df.PolarsDfT, pydantic.Tag("collected")],
        t.Annotated[_lazy_df.PolarsLazyDfT, pydantic.Tag("lazy")],
    ],
    pydantic.Discriminator(discriminator=_callable_discrim)
]
"""
An annotated type representing a union of polars DataFrame or LazyFrame.

This extra metadata allows us to include polars collected dataframes in the pydantic ecosystem, 
e.g. as a field on a model. The pydantic behavior is as follows:

* A LazyFrame instance will be kept as a LazyFrame instance, unchanged
* DataFrame instance will be parsed as a DataFrame instance, unchanged
* list of dicts will be parsed as a DataFrame. (row orientation)
* dict of lists will be parsed as a DataFrame (columnar orientation)

* serialization will return a row orientation of json-encodeable dicts, if lazy, only the first 3

Notes: this is not safe for roundtrip serialization, i.e. the a serialized instance of an instance
    of this type will not necessarily deserialize to be equal to the original instance
"""

PolarsMaybeLazyDfTypeAdapter: pydantic.TypeAdapter[PolarsMaybeLazyDfT] = pydantic.TypeAdapter(PolarsMaybeLazyDfT)
"""
type adapter for PolarsMaybeLazyDfT
"""

PolarsDfTV = t.TypeVar("PolarsDfTV", _collected_df.PolarsDfT, _lazy_df.PolarsLazyDfT)
"""common type var that could be either a lazy or collected dataframe"""



from __future__ import annotations
import typing as t
import pydantic

from spice_rack._fs_ops._fs_models import (
    _base,
    _file,
    _dir,
    _deferred
)


__all__ = (
    "FileOrDirPathT", "FileOrDirPathTypeAdapter",
    "FileOrDirDeferredPathT", "FileOrDirDeferredPathTypeAdapter"
)


def _str_parser(data: t.Any) -> t.Any:
    if isinstance(data, str):
        if data.endswith("/"):
            return _dir.DirPath.init_from_str(data)
        else:
            return _file.FilePath.init_from_str(data)

    else:
        return data


FileOrDirPathT = t.Annotated[
    _base.AbstractFileSystemObj.build_dispatched_ann(),
    pydantic.BeforeValidator(_str_parser)
]
"""extends standard dispatched type to support raw strings"""


FileOrDirPathTypeAdapter = pydantic.TypeAdapter(FileOrDirPathT)
"""gives us pydantic parsing logic outside a pydantic class"""


def _deferred_str_parser(data: t.Any) -> t.Any:
    if isinstance(data, str):
        if data.startswith("$"):
            if data.endswith("/"):
                return _deferred.DeferredDirPath.model_validate(data)
            else:
                return _deferred.DeferredFilePath.model_validate(data)
        else:
            raise ValueError(
                f"a deferred path str must start with '$', '{data}' does not"
            )
    else:
        return data


FileOrDirDeferredPathT = t.Annotated[
    t.Union[_deferred.DeferredFilePath, _deferred.DeferredDirPath],
    pydantic.BeforeValidator(_deferred_str_parser)
]
"""extends standard dispatched type to support raw strings"""


FileOrDirDeferredPathTypeAdapter: pydantic.TypeAdapter[FileOrDirDeferredPathT] = (
    pydantic.TypeAdapter(FileOrDirDeferredPathT)
)
"""gives us pydantic parsing logic outside a pydantic class"""



from __future__ import annotations
from abc import abstractmethod
import typing as t
from pathlib import Path
import pydantic

from spice_rack._fs_ops._path_strs import _base, _path_checkers
from spice_rack._fs_ops import _file_info, _exceptions


__all__ = (
    "RelFilePathStr",
    "RelDirPathStr",
    "FileOrDirRelPathT",
    "FileOrDirRelPathTypeAdapter"
)


class _AbstractRelPathStr(_base.AbstractPathStr):
    """Base class for a relative file path and relative dir path string classes"""
    @abstractmethod
    def get_name(self) -> str:
        ...

    @classmethod
    def _check_str_val(cls, __raw_str: str) -> t.List[str]:
        issues = super()._check_str_val(__raw_str)
        issues.extend(_path_checkers.rel_validator(__raw_str))
        return issues

    def get_parent(self) -> RelDirPathStr:
        parent_path_raw = str(Path(str(self)).parent)
        if not parent_path_raw.endswith("/"):
            parent_path_raw = parent_path_raw + "/"
        return RelDirPathStr(parent_path_raw)


@t.final
class RelFilePathStr(_AbstractRelPathStr):
    """
    This represents an absolute file path string. It must start with a backslash and must not end
    with a backslash.
    """
    @classmethod
    def _check_str_val(cls, __raw_str: str) -> t.List[str]:
        issues = super()._check_str_val(__raw_str)
        issues.extend(_path_checkers.file_validator(__raw_str))
        return issues

    def get_name(self, include_suffixes: bool = False) -> str:
        """get simple name, i.e. the most terminal chunk in the path"""
        name_w_suffixes = str(self).split("/")[-1]
        if include_suffixes:
            return name_w_suffixes
        else:
            return name_w_suffixes.split(".")[0]

    def get_suffixes(self) -> list[_file_info.FileExt]:
        suffixes = str(self).split(".")[1:]
        return [
            _file_info.FileExt(suffix) for suffix in suffixes
        ]

    def get_file_ext(self) -> t.Optional[_file_info.FileExt]:
        suffixes = self.get_suffixes()
        if suffixes:
            return suffixes[-1]
        else:
            return None

    def get_mime_type(self) -> t.Optional[_file_info.MimeType]:
        file_ext = self.get_file_ext()
        if file_ext:
            mime_type_maybe = file_ext.get_mime_type()
            return mime_type_maybe
        else:
            return None

    def ensure_correct_file_ext(
            self,
            choices: list[_file_info.FileExt]
    ) -> None:
        """
        ensure the file path has one of the specified extensions

        Args:
            choices: the valid extensions

        Returns: Nothing

        Raises:
            FilePathInvalidException: if the file path has no extension of it isn't
                one of the choices
        """
        file_ext = self.get_file_ext()
        if file_ext is None or file_ext not in choices:
            raise _exceptions.InvalidFileExtensionException(
                path=self,
                file_ext_found=file_ext,
                file_ext_choices=choices,
            )
        return

    def ensure_correct_mime_type(
            self,
            choices: list[_file_info.MimeType]
    ) -> None:
        """
        ensure the file path is one of the specified mime types
        Args:
            choices: the valid mime types

        Returns: Nothing

        Raises:
            InvalidFileMimeTypeException: if we cannot determine the mime type, or it is not one
                of the specified choices
        """

        mime_type_found = self.get_mime_type()
        if not mime_type_found or mime_type_found not in choices:
            raise _exceptions.InvalidFileMimeTypeException(
                path=self,
                mime_type_found=mime_type_found,
                mime_type_choices=choices,
                verbose=True
            )


@t.final
class RelDirPathStr(_AbstractRelPathStr):
    """subclasses string to enforce constraints on the dir path str"""

    @classmethod
    def _check_str_val(cls, __raw_str: str) -> t.List[str]:
        issues = super()._check_str_val(__raw_str)
        issues.extend(_path_checkers.dir_validator(__raw_str))
        return issues

    @classmethod
    def _format_str_val(cls, root_data: str) -> str:
        root_data = super()._format_str_val(root_data)
        if not root_data.endswith("/"):
            root_data = root_data + "/"
        return root_data

    def get_name(self) -> str:
        """get simple name, i.e. the most terminal chunk in the path"""
        return str(self).split("/")[-2] + "/"

    @t.overload
    def joinpath(self, rel_path: RelDirPathStr) -> RelDirPathStr:
        ...

    @t.overload
    def joinpath(self, rel_path: RelFilePathStr) -> RelFilePathStr:
        ...

    @t.overload
    def joinpath(self, rel_path: str) -> FileOrDirRelPathT:
        ...

    def joinpath(
            self,
            rel_path: t.Union[str, FileOrDirRelPathT]
    ) -> FileOrDirRelPathT:
        rel_path_parsed = FileOrDirRelPathTypeAdapter.validate_python(
            rel_path
        )
        extended_path_raw: str = str(self) + str(rel_path_parsed)

        if isinstance(rel_path_parsed, RelFilePathStr):
            extended_path = RelFilePathStr(extended_path_raw)

        elif isinstance(rel_path_parsed, RelDirPathStr):
            extended_path = RelDirPathStr(extended_path_raw)

        else:
            raise TypeError(f"{type(rel_path_parsed)} unexpected")

        return extended_path


_TagsT = t.Literal["dir", "file"]


def _discrim(raw_str: str) -> _TagsT:
    if raw_str.endswith("/"):
        return "dir"
    else:
        return "file"


FileOrDirRelPathT = t.Annotated[
    t.Union[
        t.Annotated[RelFilePathStr, pydantic.Tag("file")],
        t.Annotated[RelDirPathStr, pydantic.Tag("dir")],
    ],
    pydantic.Discriminator(_discrim)
]


FileOrDirRelPathTypeAdapter: pydantic.TypeAdapter[FileOrDirRelPathT] = pydantic.TypeAdapter(
    FileOrDirRelPathT
)



from __future__ import annotations
import typing as t
import sys
from pydantic import Field

from spice_rack._logging._sinks import _base

if t.TYPE_CHECKING:
    from loguru import Logger

__all__ = (
    "SysLogSink",
)


class SysLogSink(_base.AbstractLogSink, class_id="sys"):
    channel: t.Literal["stdout", "stderr"] = "stdout"
    struct_log: bool = Field(
        description="if true, we serialize the logs for usage in a structured way. "
                    "this field is used when building the loguru_kwargs "
                    "passed in when setting up the sink."
                    "This must be False for system-based log sink",
        default=False,
    )

    def setup(self, logger: Logger, **custom_loguru_kwargs) -> None:
        std_dest = getattr(sys, self.channel)

        loguru_kwargs = self._get_loguru_kwargs()
        loguru_kwargs.update(custom_loguru_kwargs)
        assert std_dest
        logger.add(std_dest, **loguru_kwargs)



from __future__ import annotations
import pydantic

from spice_rack._bases import _base_base

__all__ = (
    "VALUE_MODEL_CONFIG",
    "ValueModelBase"
)

_value_config = pydantic.ConfigDict(
    frozen=True,
    extra="forbid",
)
VALUE_MODEL_CONFIG: pydantic.ConfigDict = {
    **_base_base.BASE_MODEL_CONFIG,
    **_value_config
}


class ValueModelBase(_base_base.PydanticBase):
    """a frozen model"""
    model_config = VALUE_MODEL_CONFIG



from __future__ import annotations
from typing import Literal
from typing_extensions import TypeAlias

from pydantic import RootModel


__all__ = (
    "EMPTY",
    "EmptyT"
)


class _Empty(RootModel):
    """
    class to represent a null, always used as a default.
    This is validated as a pydantic field annotation, or within one.
    """
    root: Literal[
        "EMPTY_0c616003538843f2a01f9d6029a3ea65"
    ] = "EMPTY_0c616003538843f2a01f9d6029a3ea65"

    def __hash__(self) -> int:
        return hash("EMPTY_0c616003538843f2a01f9d6029a3ea65")


EMPTY = _Empty()

EmptyT: TypeAlias = _Empty



from __future__ import annotations
import typing as t
import gcsfs  # noqa
import pydantic

from spice_rack._fs_ops import _path_strs, _helpers
from spice_rack._fs_ops._file_systems import _base
from spice_rack import _gcp_auth


__all__ = (
    "GcsFileSystem",
)


@t.final
class GcsFileSystem(_base.AbstractFileSystem, class_id="gcs"):
    """wrapper for the GCSFileSystem file system"""
    creds: _gcp_auth.AnyGcpAuthStrat = pydantic.Field(
        description="the credentials we use to authenticate to the gcs bucket",
        default_factory=_gcp_auth.AnyGcpAuthStrat.init_default
    )

    @classmethod
    def get_fs_specific_prefix(cls) -> str:
        return "gs://"

    def _get_gcsfs_token(
            self,
    ) -> t.Union[str, dict]:
        creds: _gcp_auth.auth_strategies.AbstractGcpAuthStrategy = self.creds.root

        if isinstance(creds, _gcp_auth.auth_strategies.AnonAuthStrategy):
            return "anon"

        elif isinstance(creds, _gcp_auth.auth_strategies.DefaultAuthStrategy):
            if creds.default_creds_found():
                return "google_default"
            else:
                if creds.on_no_default == "use_anon":
                    return "anon"
                elif creds.on_no_default == "raise":
                    raise ValueError(
                        "no default gcp credentials found in the environment"
                    )
                else:
                    raise ValueError(
                        f"'{creds.on_no_default}' is unexpected value for 'on_no_default'"
                    )
        elif isinstance(creds, _gcp_auth.auth_strategies.ServiceAcctKeyFileAuthStrategy):
            return creds.file_path
        elif isinstance(creds, _gcp_auth.auth_strategies.ServiceAcctKeyAuthStrategy):
            return creds.key_data.dict(exclude_defaults=True)
        else:
            raise ValueError(f"unexpected auth strategy type: {type(self.creds,)}")

    def build_fsspec_file_system(self) -> gcsfs.GCSFileSystem:
        token = self._get_gcsfs_token()
        return gcsfs.GCSFileSystem(
            token=token
        )

    @pydantic.validate_call
    def make_dir(
            self,
            __path: _path_strs.AbsoluteDirPathStr,
            *,
            if_exists: t.Literal["raise", "return"] = "return",
            create_parents: bool = True
    ) -> None:

        # creating cloud dir doesn't work like local bc how
        # they treat directories. We create a placeholder
        # file when making a directory to imitate this.
        super().make_dir(__path, if_exists=if_exists, create_parents=create_parents)
        if not self.exists(__path):
            placeholder_path = __path.joinpath(rel_path=_helpers.get_placeholder_rel_path())
            with self.open_file(placeholder_path, "wb") as f:
                f.write("placeholder text".encode())

    @pydantic.validate_call
    def iter_dir_contents(
            self,
            __path: _path_strs.AbsoluteDirPathStr,
    ) -> t.Iterator[_path_strs.FileOrDirAbsPathT]:
        """
        same iter as base class, except we also skip the file if it is the placeholder file
        """
        for path_i in super().iter_dir_contents(__path):
            if _helpers.is_placeholder_file_path(path_i):
                continue
            else:
                yield path_i

    def get_home_dir(self) -> _path_strs.AbsoluteDirPathStr:
        return _path_strs.AbsoluteDirPathStr("/")



from __future__ import annotations
import typing as t
import pydantic

__all__ = (
    "ColumnarDictT", "ColumnarDictTypeAdapter",
    "RowDictsT", "RowDictsTypeAdapter",

    "RowDictsJsonDumpedT", "RowDictsJsonDumpedTypeAdapter",
    "ColumnarDictJsonDumpedT", "ColumnarDictJsonDumpedTypeAdapter"

)


ColumnarDictT = t.Dict[str, t.List[t.Any]]
ColumnarDictTypeAdapter = pydantic.TypeAdapter(ColumnarDictT)

RowDictsT = t.List[t.Dict[str, t.Any]]
RowDictsTypeAdapter = pydantic.TypeAdapter(RowDictsT)

RowDictsJsonDumpedT = t.List[t.Dict[str, pydantic.JsonValue]]
RowDictsJsonDumpedTypeAdapter = pydantic.TypeAdapter(RowDictsJsonDumpedT)

ColumnarDictJsonDumpedT = t.Dict[str, t.List[pydantic.JsonValue]]
ColumnarDictJsonDumpedTypeAdapter = pydantic.TypeAdapter(ColumnarDictJsonDumpedT)



from __future__ import annotations
import typing as t

from pydantic import PrivateAttr
from pydantic.fields import FieldInfo
from pydantic_settings import BaseSettings
from spice_rack._bases._settings._sources import _base

if t.TYPE_CHECKING:
    from spice_rack._bases._settings._base import SettingsBase


__all__ = (
    "SingletonSource",
)


SettingsInstTV = t.TypeVar("SettingsInstTV", bound=BaseSettings)


class SingletonSource(_base.CustomSettingSourceBase, t.Generic[SettingsInstTV]):
    """stores prior reads and uses them if available"""
    settings_cls: t.Type[BaseSettings]
    _registry: _Registry[SettingsInstTV] = PrivateAttr()

    def __init__(self, settings_cls: t.Type[SettingsInstTV]):
        super().__init__(settings_cls=settings_cls)

        settings_cls: t.Type[BaseSettings] = self.settings_cls  # type: ignore
        self._registry = _Registry[settings_cls]()  # type: ignore

    def get_field_value(self, field: FieldInfo, field_name: str) -> tuple[t.Any, str, bool]:

        # copied this from the implementation for another settings_base source in pydantic_settings
        # Nothing to do here. Only implement the return statement to make mypy happy
        return None, '', False

    def __call__(self) -> dict[str, t.Any]:
        cached_inst_maybe = self._registry.get_maybe(self.settings_cls)  # type: ignore
        if cached_inst_maybe:
            return cached_inst_maybe.model_dump()
        else:
            return {}

    def add_to_registry(self, inst: SettingsInstTV) -> None:
        self._registry.cache(inst)


class _Registry(t.Generic[SettingsInstTV]):
    def __init__(self):
        self._registry: dict[t.Type[SettingsInstTV], SettingsInstTV] = {}    # type: ignore

    def get_maybe(
            self,
            settings_cls: t.Type[SettingsInstTV]
    ) -> t.Optional[SettingsBase]:
        return self._registry.get(settings_cls)  # type: ignore

    def cache(
            self,
            settings_inst: SettingsInstTV
    ) -> None:
        self._registry[type(settings_inst)] = settings_inst



from __future__ import annotations
import typing as t
import pydantic
import pydantic_core
import polars as pl

from spice_rack._polars_service._types import _dicts, _discrim_helper


__all__ = (
    "PolarsDfT",
    "PolarsDfTypeAdapter"
)


class _PolarsDfPydanticAnnotation:
    @classmethod
    def _get_df_schema(cls) -> pydantic_core.CoreSchema:
        """already a df"""
        return pydantic_core.core_schema.is_instance_schema(
            cls=pl.DataFrame,
            cls_repr="pl.DataFrame",
            ref="polars_dataframe",
        )

    @classmethod
    def _get_lazy_df_schema(cls) -> pydantic_core.CoreSchema:
        return pydantic_core.core_schema.chain_schema(
            [
                pydantic_core.core_schema.is_instance_schema(
                    cls=pl.LazyFrame,
                    cls_repr="pl.LazyFrame",
                    ref="polars_lazyframe",
                ),
                pydantic_core.core_schema.no_info_plain_validator_function(
                    function=lambda _lazy_df: _lazy_df.collect()
                )
            ]
        )

    @classmethod
    def _get_row_dicts_schema(cls) -> pydantic_core.CoreSchema:
        return pydantic_core.core_schema.chain_schema(
            [
                _dicts.RowDictsTypeAdapter.core_schema,
                pydantic_core.core_schema.no_info_plain_validator_function(
                    function=lambda _rows: pl.DataFrame(_rows)
                ),
            ]
        )

    @classmethod
    def _get_columnar_dict_schema(cls) -> pydantic_core.CoreSchema:
        return pydantic_core.core_schema.chain_schema(
            [
                _dicts.ColumnarDictTypeAdapter.core_schema,
                pydantic_core.core_schema.no_info_plain_validator_function(
                    function=lambda _column_data: pl.DataFrame(_column_data)
                ),
            ]
        )

    @classmethod
    def _get_core_json_schema(cls) -> pydantic_core.CoreSchema:
        """
        return the schema we use for json serialization and deserialization.
        """
        return _dicts.RowDictsJsonDumpedTypeAdapter.core_schema

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: t.Any,
        _handler: pydantic.GetCoreSchemaHandler,
    ) -> pydantic_core.core_schema.CoreSchema:
        """
        creates a pydantic core schema that behaves in the following ways

        * DataFrame instance will be parsed as a DataFrame instance, unchanged
        * LazyFrame instance will be collected to be a DataFrame.
        * list of dicts will be parsed as a DataFrame. (row orientation)
        * dict of lists will be parsed as a DataFrame (columnar orientation)

        * serialization will return a row orientation of json-encodeable dicts

        Args:
            _source_type: the class calling this, not sure exactly what it means beyond that,
                pydantic expects it.
            _handler: another param pydantic expected

        Returns:
            pydantic_core.core_schema.CoreSchema
        """
        from spice_rack._polars_service import _services

        choices = _discrim_helper.build_choices(
            df_schema=cls._get_df_schema(),
            lazy_df_schema=cls._get_lazy_df_schema(),
            row_dicts_schema=cls._get_row_dicts_schema(),
            columnar_dict_schema=cls._get_columnar_dict_schema()
        )

        return pydantic_core.core_schema.json_or_python_schema(
            python_schema=pydantic_core.core_schema.tagged_union_schema(
              choices=choices, discriminator=_discrim_helper.discrim_func
            ),
            serialization=pydantic_core.core_schema.plain_serializer_function_ser_schema(
                lambda _df: _services.get_json_safe_row_dicts(_df),
                when_used="json-unless-none"
            ),
            json_schema=cls._get_core_json_schema(),
        )

    @classmethod
    def __get_pydantic_json_schema__(
            cls,
            _core_schema: pydantic_core.core_schema.CoreSchema,
            handler: pydantic.GetJsonSchemaHandler
    ):
        """
        copy the row schema
        """
        return handler(cls._get_core_json_schema())


PolarsDfT = t.Annotated[
    pl.DataFrame,
    _PolarsDfPydanticAnnotation
]
"""
An annotated type representing a polars DataFrame.

This extra metadata allows us to include polars collected dataframes in the pydantic ecosystem, 
e.g. as a field on a model. The pydantic behavior is as follows:

* DataFrame instance will be parsed as a DataFrame instance, unchanged
* LazyFrame instance will be collected to be a DataFrame.
* list of dicts will be parsed as a DataFrame. (row orientation)
* dict of lists will be parsed as a DataFrame (columnar orientation)

* serialization will return a row orientation of json-encodeable dicts

Notes: this is not safe for roundtrip serialization, i.e. the a serialized instance of an instance
    of this type will not necessarily deserialize to be equal to the original instance
"""

PolarsDfTypeAdapter: pydantic.TypeAdapter[PolarsDfT] = pydantic.TypeAdapter(PolarsDfT)
"""
type adapter for PolarsDfT
"""



from spice_rack._bases._exception._exception_base import *
from spice_rack._bases._exception._exception_exception import *
from spice_rack._bases._exception._error_info import *
from spice_rack._bases._exception._external_wrapper import *
from spice_rack._bases._exception import _http as http



from spice_rack._frozen_registry._exceptions._dupe_keys import *
from spice_rack._frozen_registry._exceptions._item_ix_invalid import *
from spice_rack._frozen_registry._exceptions._key_already_exists import *
from spice_rack._frozen_registry._exceptions._missing_key import *



from __future__ import annotations
import typing as t
import pydantic
import pydantic_core

from spice_rack import _bases, _logging


__all__ = (
    "MutuallyExclusiveFieldsException",
    "MutuallyExclusiveFieldsValidator"
)


class _MutuallyExclusiveFieldsErrorInfo(_bases.exceptions.ErrorInfoBase):
    class_name: str = pydantic.Field(
        description="the name of the class we are validating"
    )
    field_in_group: t.List[str] = pydantic.Field(
        description="the list of all field names we are grouping together and enforcing to be mutually exclusive"
    )
    fields_missing: t.List[str] = pydantic.Field(description="fields specified to check, but not found on the model")
    fields_specified: t.List[str] = pydantic.Field(description="multiple fields specified, when only one should be")


class MutuallyExclusiveFieldsException(_bases.exceptions.CustomExceptionBase[_MutuallyExclusiveFieldsErrorInfo]):
    def __init__(
            self,
            detail: str,
            class_name: str,
            field_in_group: t.List[str],
            fields_missing: t.List[str],
            fields_specified: t.List[str],

    ):
        error_info = {
            "class_name": class_name,
            "field_in_group": field_in_group,
            "fields_missing": fields_missing,
            "fields_specified": fields_specified
        }
        super().__init__(detail=detail, error_info=error_info, extra_info=None, verbose=True)


class MutuallyExclusiveFieldsValidator:
    """validates the specified fields are mutually exclusive"""
    def __init__(
            self,
            field_names: t.List[str],
            model_cls: t.Optional[t.Type[pydantic.BaseModel]] = None,
            at_least_one: bool = True,
            check_field_names: bool = True
    ):
        self.field_names = field_names
        self.at_least_one = at_least_one
        self.check_field_names = check_field_names

        if model_cls:
            missing_field_names: t.List[str] = []
            for field_name in self.field_names:
                if field_name not in model_cls.model_fields:
                    missing_field_names.append(field_name)
            if self.check_field_names and missing_field_names:
                raise ValueError(
                    f"error setting up '{self.__class__.__name__}', the model class, "
                    f"'{model_cls.__name__}', is missing"
                    f" the following fields: {missing_field_names}"
                )
            else:
                _logging.Logger.get_logger().info(
                    f"'{self.__class__.__name__}' configured on the '{model_cls.__name__}' class "
                    f" but the following fields weren't found on the class: {missing_field_names}"
                )

    def validate_inst(self, model_inst: pydantic.BaseModel) -> pydantic.BaseModel:
        """
        performs the actual validation
        Args:
            model_inst: the instance we are validating

        Returns: the instance passed in, unchanged

        Raises:
            MutuallyExclusiveFieldsException: if we find an error. todo: add more info on why we'd raise an error
        """
        data_dict = model_inst.model_dump(
            mode="python", include=set(self.field_names)
        )

        missing_field_names: t.List[str] = []
        specified_field_names: t.List[str] = []

        for field_name in self.field_names:
            if field_name not in data_dict:
                missing_field_names.append(field_name)
            if data_dict.get(field_name) is not None:
                specified_field_names.append(field_name)

        if self.check_field_names and missing_field_names:
            raise MutuallyExclusiveFieldsException(
                detail="not every field in the specified field names was specified on the model",
                class_name=model_inst.__class__.__name__,
                fields_missing=missing_field_names,
                fields_specified=specified_field_names,
                field_in_group=self.field_names,
            )

        if self.at_least_one and not specified_field_names:
            raise MutuallyExclusiveFieldsException(
                detail="at least one of the fields in the group should be non-null. All were found to be null",
                class_name=model_inst.__class__.__name__,
                fields_missing=missing_field_names,
                fields_specified=specified_field_names,
                field_in_group=self.field_names,
            )

        if len(specified_field_names) > 1:
            raise MutuallyExclusiveFieldsException(
                detail="found values for more than one field in a group that is specified to be mutually exclusive",
                class_name=model_inst.__class__.__name__,
                fields_missing=missing_field_names,
                fields_specified=specified_field_names,
                field_in_group=self.field_names,
            )
        return model_inst

    def _get_func(self) -> pydantic.functional_validators.ModelAfterValidatorWithoutInfo:

        def func(model_inst: pydantic.BaseModel) -> pydantic.BaseModel:
            try:
                return self.validate_inst(model_inst)
            except MutuallyExclusiveFieldsException as e:
                raise e.as_pydantic_error()
            except Exception as e:
                raise e

        return func

    def __get_pydantic_core_schema__(
            self,
            __source: t.Any,
            __handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.CoreSchema:
        val_func = self._get_func()

        schema = __handler(__source)
        schema = pydantic_core.core_schema.no_info_after_validator_function(
            function=val_func,
            schema=schema
        )
        return schema



from __future__ import annotations
import typing as t

if t.TYPE_CHECKING:
    from spice_rack._fs_ops import _file_systems


__all__ = (
    "infer_file_system",
)


def infer_file_system(raw_path: str) -> _file_systems.AnyFileSystemT:
    from spice_rack._fs_ops import _file_systems

    # if starts with "gs://" we assume gcp
    if raw_path.startswith("gs://"):
        return _file_systems.GcsFileSystem()

    else:
        # fallback is local
        return _file_systems.LocalFileSystem()
